"""OSP-utilities for arcp-iris"""

from urllib.parse import quote, urlencode

from arcp import arcp_name, arcp_random, is_arcp_uri
from arcp.parse import ARCPParseResult, parse_arcp

from osp.core.cuds import Cuds


def check_arcp(cuds: Cuds) -> bool:
    """Check whether the iri of a Cuds is an arcp-uri"""
    return is_arcp_uri(cuds.iri)


def wrap_arcp(cuds: Cuds) -> ARCPParseResult:
    """Parse the iri of a Cuds according to the arcp-rules"""
    return parse_arcp(cuds.iri)


def make_remote(*args, **kwargs) -> str:
    """Make a named arcp-iri"""
    return arcp_name(*args, **kwargs)


def make_local(*args, **kwargs) -> str:
    """Make a randomized arcp-iri"""
    return arcp_random(*args, **kwargs)


def make_arcp(*args, query: dict = None, dtype: str = "local", **kwargs) -> str:
    """Make a arcp-iri, named or randomized."""
    if query:
        for key, value in query.items():
            query[key] = [quote(".".join(item)) for item in value]
        query = urlencode(query, doseq=True)
    if dtype == "local":
        func = make_local
    elif dtype == "remote":
        func = make_remote
    else:
        raise ValueError(
            f"Unknown conversion-type {dtype}. Must be 'local' or 'remote'."
        )
    return func(*args, **kwargs, query=query)



"""SimPhoNy-Celery Wrapper"""

from .celery_workflow_engine import CeleryWorkflowEngine
from .celery_workflow_wrapper import CeleryWorkflowSession

__all__ = ["CeleryWorkflowEngine", "CeleryWorkflowSession"]



"""Module for injection of dependencies into fastapi-endpoints and routers"""

import logging
import os
import tempfile
from importlib import import_module
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional
from uuid import UUID

from celery import Celery
from celery.app.control import Inspect
from fastapi import Depends, Query, UploadFile
from minio import Minio
from shieldapi.frameworks.fastapi import AuthTokenBearer
from urllib3.response import HTTPResponse

from osp.settings import AppConfig
from osp.utilities.helper import _get_download, _get_upload

if TYPE_CHECKING:  # pragma: no cover
    from typing import Collection


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def get_app() -> "Celery":
    """Return celery app"""
    from .tasks import celery  # pylint: disable=import-outside-toplevel

    return celery


def get_method() -> Callable:
    """Return celery-tasks for running simulation"""
    from .tasks import run_simulation  # pylint: disable=import-outside-toplevel

    return run_simulation


def get_appconfig() -> AppConfig:
    """Returns app-config."""
    return AppConfig()


def get_dependencies() -> "List[Depends]":
    """Return dependencies for FastAPI-middleware for authentication"""
    app_config = get_appconfig()

    if app_config.authentication_dependencies:
        dependencies = [Depends(AuthTokenBearer(bearerFormat="JWT"))]
    else:
        dependencies = []
        logger.info("No dependencies for authentication assigned.")
    return dependencies


def get_models() -> "Dict[str, Callable]":
    """Get registry of pydantic models."""
    app_config = AppConfig()
    modules = [module.strip().split(":") for module in app_config.schemas.split("|")]
    return {
        classname: getattr(import_module(module), classname)
        for (module, classname) in modules
    }


def depends_modellist() -> "List[str]":
    """Get list of model names."""
    registry = get_models()
    return list(registry.keys())


def depends_inspect(celery_app: "Celery" = Depends(get_app)) -> Inspect:
    """Return the object from the celery app for the backend inspection."""
    return celery_app.control.inspect()


def depends_registry(
    inspect: Inspect = Depends(depends_inspect),
) -> "Dict[Any, Collection[str]]":
    """Return a mapping between the workers id and registered tasks"""
    registry = inspect.registered_tasks()
    if not registry:
        registry = {}
    return registry


def depends_tasks(registry: dict = Depends(depends_registry)) -> "List[str]":
    """Return the list of tasks available from the message broker."""
    task_list: List[str] = []
    if not registry:
        return task_list
    for tasks in registry.values():
        task_list += tasks
    return task_list


def depends_minio(settings: AppConfig = Depends(get_appconfig)) -> Minio:
    """Instantiate MinIO-client via `Depends`."""
    return Minio(
        settings.minio_endpoint,
        access_key=settings.minio_user.get_secret_value(),
        secret_key=settings.minio_password.get_secret_value(),
        secure=False,
    )


def depends_upload(
    file: UploadFile,
    uuid: Optional[UUID] = None,
    minio_client: Minio = Depends(depends_minio),
) -> UUID:
    """Upload file with minio client and return upload-id via `Depends`."""
    suffix = os.path.splitext(file.filename)[-1]
    with tempfile.NamedTemporaryFile(suffix=suffix, delete=False) as temp:
        temp.write(file.file.read())
    return _get_upload(temp.name, uuid, minio_client)


def depends_upload_enabled(config: AppConfig = Depends(get_appconfig)) -> bool:
    """Check whether direct upoad to data cache is enabled or not"""
    return config.enable_upload


def depends_download(
    dataset_name: str = Query(..., title="Cache ID received after the upload."),
    minio_client: Minio = Depends(depends_minio),
) -> HTTPResponse:
    """Return `HTTPResponse` from uuid through minio client via `Depends`."""
    return _get_download(dataset_name, minio_client)


def depends_logs(
    transformation_id: str = Query(..., title="task id of the submitted job"),
    minio_client: Minio = Depends(depends_minio),
) -> HTTPResponse:
    """Return `HTTPResponse` from uuid through minio client via `Depends`."""
    return _get_download(transformation_id, minio_client)



"""Exceptions for osp-utilities"""


class MinioDownloadError(Exception):
    """Error while fetching object from MinIO-instance."""


class MinioConnectionError(Exception):
    """Error while connecting to MinIO-instance."""



"""Main FastAPI-middleware for running remote celery tasks"""
import logging
import os
from datetime import datetime
from typing import Annotated, Any, Dict, List, Optional, Union
from uuid import UUID

import pkg_resources
import uvicorn
from celery import Celery
from fastapi import Body, Depends, FastAPI, HTTPException, Query, Response
from fastapi.openapi.utils import get_openapi
from fastapi.responses import JSONResponse, RedirectResponse

# from fastapi.responses import StreamingResponse
from fastapi_plugins import (
    config_plugin,
    get_config,
    redis_plugin,
    register_config,
    register_middleware,
)
from minio import ServerError
from pydantic.error_wrappers import ValidationError
from pydantic.schema import schema
from urllib3.response import HTTPResponse

from osp.settings import AppConfig

from .dependencies import (
    depends_download,
    depends_logs,
    depends_modellist,
    depends_registry,
    depends_upload,
    depends_upload_enabled,
    get_app,
    get_appconfig,
    get_dependencies,
    get_models,
)
from .models import (
    InfoType,
    RegisteredModels,
    RegisteredTaskModel,
    SubmissionBody,
    TaskCreateModel,
    TaskKillModel,
    TaskResultModel,
    TaskStatusModel,
    TransformationStatus,
    UploadDataResponse,
    UploadNotEnabledError,
    task_to_transformation_map,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
PACKAGE_NAME = "reaxpro-workflow-service"


app = register_middleware(FastAPI(dependencies=get_dependencies()))
register_config(AppConfig)
config = get_config()


def custom_openapi():
    """Overwrite openapi schema"""
    if app.openapi_schema:
        return app.openapi_schema
    openapi_schema = get_openapi(
        title="ReaxPro Workflow Service",
        description="Multi-scale simulation platform for reactive processes",
        version=pkg_resources.get_distribution(PACKAGE_NAME).version,
        contact={
            "name": "Matthias Büschelberger",
            "url": "https://materials-marketplace.eu/",
            "email": "matthias.bueschelberger@iwm.fraunhofer.de",
        },
        license_info={
            "name": "BSD-3-clause",
            "url": "https://opensource.org/licenses/bsd-3-clause",
        },
        servers=[{"url": config.external_hostname}]
        if config.external_hostname
        else config.external_hostname,
        routes=app.routes,
    )
    openapi_schema["info"]["x-api-version"] = "0.4.0"
    openapi_schema["info"]["x-products"] = [{"name": "Monthly"}]
    app.openapi_schema = openapi_schema
    return app.openapi_schema


app.openapi = custom_openapi


@app.get("/", include_in_schema=False)
async def root():
    """Redirect to /docs page"""
    return RedirectResponse(url="/docs")


@app.put("/data/cache", operation_id="createDataset")
async def upload_data(
    object_key: UUID = Depends(depends_upload),
    upload_enabled: bool = Depends(depends_upload_enabled),
) -> UploadDataResponse:
    """Upload data from internal cache"""
    if not upload_enabled:
        raise UploadNotEnabledError("Direct upload to cache is not enabled")
    return UploadDataResponse(id=object_key, last_modified=str(datetime.now()))


@app.get("/data/cache/{dataset_name}", operation_id="getDataset")
async def download_data(
    dataset_name: str,
    response=Depends(depends_download),
) -> Response:
    """Download file via StreamingResponse"""
    filename = str(dataset_name) + response.headers.get("x-amz-meta-suffix")
    # return StreamingResponse(
    #     iter(response.readlines()),
    #     media_type="multipart/form-data",
    #     headers={"Content-Disposition": f"attachment; filename={filename}"},
    # )
    response = Response(response.data)
    # Set the appropriate headers
    response.headers["Content-Disposition"] = f'attachment; filename="{filename}"'
    response.headers["Content-Type"] = "application/octet-stream"
    return response


@app.get("/workers/registered")
async def get_workers_available(
    registry: Union[dict, None] = Depends(depends_registry)
) -> RegisteredTaskModel:
    """Return the list of runable and registered tasks"""
    return RegisteredTaskModel(
        **{
            "message": "Fetched registry of executable tasks on remote workers.",
            "registered_tasks": registry,
        }
    )


@app.get("/models", operation_id="getModels")
async def get_model_available(
    registry: List[str] = Depends(depends_modellist),
) -> RegisteredModels:
    """Return the list of runable and registered tasks"""
    return RegisteredModels(
        **{
            "message": "Fetched registry of data models.",
            "registered_models": registry,
        }
    )


@app.get("/info", operation_id="getInfo")
async def get_info(
    info_type: InfoType = Query(
        ..., title="Type of info to be retrieved. schema or example"
    ),
    model_name: Optional[str] = Query(
        ...,
        enum=depends_modellist(),
        title="Name of the data model for which info should be retrieved.",
    ),
) -> Union[Dict[Any, Any], Any]:
    """Get specific app info"""
    if info_type == InfoType.SCHEMA:
        # Retreive schema of a model registered in the app
        models = get_models()
        model = models.get(model_name)
        response = schema([model])
    elif info_type == InfoType.EXAMPLE:
        # Retrieve an example for a model registered in the app
        models = get_models()
        model = models.get(model_name)
        response = JSONResponse(model.Config.schema_extra["example"])
    return response


@app.post("/transformations", operation_id="newTransformation")
async def create_model(
    model_name: str = Query(
        ...,
        enum=depends_modellist(),
        title="Name of the data model to be instanciated.",
    ),
    body=Body(..., media_type="application/json"),
    models=Depends(get_models),
) -> TaskCreateModel:
    """Initialize transformation with respect to a certain model."""
    if "parameters" in body:
        body = body["parameters"]
    # Create a semantic model regisered in the app
    model = models.get(model_name)
    instance = model(**body)
    return TaskCreateModel(id=instance.uuid)


@app.patch("/transformations/{transformation_id}", operation_id="updateTransformation")
async def update_task(
    transformation_id: str,
    body: Annotated[
        SubmissionBody,
        Body(),
    ],
    celery_app: "Celery" = Depends(get_app),
    settings: AppConfig = Depends(get_appconfig),
) -> TaskStatusModel:
    """Get status of the transformation."""
    state = body.state
    if state == TransformationStatus.RUNNING:
        # Send a task to a remote celery worker
        task = celery_app.send_task(
            settings.worker_name,
            kwargs={"cache_key": transformation_id, "store_tarball": False},
            queue=settings.worker_name,
        )
        response = TaskStatusModel(
            status=task.status,
            state=task_to_transformation_map[task.state],
            result=task.result,
            traceback=task.traceback,
            id=task.id,
            args=task.args,
            kwargs=task.kwargs,
            date_done=task.date_done,
        )
    elif state == TransformationStatus.STOPPED:
        # Kill a submitted task with certain id
        task = celery_app.AsyncResult(transformation_id)
        if not task.date_done:
            task.revoke(terminate=True)
            message = "Killing scheduled."
        else:
            message = "Task already terminated."
        response = TaskKillModel(
            message=message,
            status=task.status,
            state=task_to_transformation_map[task.state],
            result=task.result,
            traceback=task.traceback,
            id=task.id,
            args=task.args,
            kwargs=task.kwargs,
            date_done=task.date_done,
        )
    else:
        raise HTTPException(status_code=400, detail=f"Unknown task status: {state}")
    return response


@app.get("/logs", operation_id="getLogs")
def get_logs(
    transformation_id: str = Query(  # pylint: disable=unused-argument
        ..., title="task id of the submitted job"
    ),
    response: HTTPResponse = Depends(depends_logs),
) -> Response:
    """Get logging messages from a task."""
    return Response(content=response.data, media_type="text/plain")


@app.get(
    "/transformations/{transformation_id}/state", operation_id="getTransformationState"
)
async def get_status(
    transformation_id: str = Query(..., title="task id of the job to be killed"),
    celery_app: "Celery" = Depends(get_app),
) -> TaskStatusModel:
    """Fetch the status of a submitted task with certain id"""
    task = celery_app.AsyncResult(transformation_id)
    return TaskStatusModel(
        status=task.status,
        state=task_to_transformation_map[task.state],
        traceback=task.traceback,
        id=task.id,
        args=task.args,
        kwargs=task.kwargs,
        date_done=task.date_done,
    )


@app.get("/transformations/{transformation_id}", operation_id="getTransformation")
async def get_result(
    transformation_id: str = Query(..., title="task id of the submitted job"),
    celery_app: "Celery" = Depends(get_app),
) -> TaskResultModel:
    """Return the results for submitted task with a certain id"""
    result = celery_app.AsyncResult(transformation_id)
    if not result.ready():
        raise HTTPException(status_code=400, detail="Task is not ready yet.")
    return TaskResultModel(
        parameters=result.result,
        id=transformation_id,
        traceback=result.traceback,
        date_done=result.date_done,
    )


@app.exception_handler(ValidationError)
async def validation_exception_handler(request, exc):  # pylint: disable=unused-argument
    """Return response based on the error"""
    # Convert the error into a list of error messages
    error_messages = []
    for error in exc.errors():
        error_messages.append(f"Field '{':'.join(error['loc'])}' {error['msg']}")

    # Raise an HTTPException with a 422 status code and the error messages
    return JSONResponse(status_code=422, content=error_messages)


# @app.exception_handler(ValueError)
# async def value_error_handler(request, exc):  # pylint: disable=unused-argument
#     """Return response based on the error"""
#     # Raise an HTTPException with a 422 status code and the error messages
#     return JSONResponse(status_code=422, content=list(exc.args))


@app.exception_handler(ServerError)
async def download_error_handler(request, exc):  # pylint: disable=unused-argument
    """Return response based on the MinioDownloadError"""
    # Raise an HTTPException with a 422 status code and the error messages
    return JSONResponse(status_code=422, content="Error while fetching the resource")


@app.exception_handler(UploadNotEnabledError)
async def upload_error_handler(request, exc):  # pylint: disable=unused-argument
    """Return response based on the MinioDownloadError"""
    # Raise an HTTPException with a 422 status code and the error messages
    return JSONResponse(
        status_code=403,
        content="""Direct upload to cache is not enabled.
    Please contact an administrator.""",
    )


# @app.exception_handler(MinioConnectionError)
# async def connection_error_handler(request, exc):  # pylint: disable=unused-argument
#     """Return response based on the MinioDownloadError"""
#     # Raise an HTTPException with a 422 status code and the error messages
#     return JSONResponse(
#         status_code=404, content="The requested resource does not exist."
#     )


@app.on_event("startup")
async def on_startup() -> None:
    """Define functions for app during startup"""
    await config_plugin.init_app(app, config)
    await config_plugin.init()
    await redis_plugin.init_app(app, config=config)
    await redis_plugin.init()


@app.on_event("shutdown")
async def on_shutdown() -> None:
    """Define functions for app during shutdown"""
    await redis_plugin.terminate()
    await config_plugin.terminate()


if __name__ == "__main__":
    host = os.environ["REAXPRO_FASTAPI_HOST"]
    port = int(os.environ["REAXPRO_FASTAPI_PORT"])
    uvicorn.run(app, host=host, port=port, log_level="debug")



"""MinIO-related utilities for SimPhoNy-OSP"""
from minio import Minio

from osp.settings import AppConfig


def get_minio() -> Minio:
    """Instantiate MinIO-client without `Depends`."""
    settings = AppConfig()
    return Minio(
        settings.minio_endpoint,
        access_key=settings.minio_user.get_secret_value(),
        secret_key=settings.minio_password.get_secret_value(),
        secure=False,
    )



"""S3Handler for logging of celery-tasks"""
import logging

from osp.utilities import get_boto3, get_minio


class S3Handler(logging.Handler):
    """Logging handler related to the S3-client"""

    def __init__(self, task_id, s3_client):
        """Initalize the class with the bucket name and key."""
        super().__init__()
        self._task_id = task_id
        self._s3_client = s3_client

    def emit(self, record):
        """Emit the record and push object to minio via S3-client."""
        log_entry = self.format(record)
        try:
            existing_object = self._s3_client.get_object(
                Bucket=self._task_id, Key=self._task_id
            )
            existing_logs = existing_object["Body"].read().decode()
        except Exception:
            existing_logs = ""

        updated_logs = existing_logs + log_entry + "\n"

        self._s3_client.put_object(
            Bucket=self._task_id,
            Key=self._task_id,
            Body=updated_logs.encode(),
        )


def get_s3handler(task_id: str) -> S3Handler:
    """Get S3 handler for logging."""
    minio_client = get_minio()
    if not minio_client.bucket_exists(task_id):
        minio_client.make_bucket(task_id)
    s3_client = get_boto3()
    return S3Handler(
        task_id,
        s3_client,
    )



"""SimPhoNy-wrapper for celery-workflows"""

import logging
from typing import TYPE_CHECKING

from osp.core.namespaces import emmo
from osp.core.session import SimWrapperSession

from .celery_workflow_engine import CeleryWorkflowEngine

if TYPE_CHECKING:
    from typing import UUID, Any, Dict, List, Optional

    from pydantic import BaseSettings

    from osp.core.cuds import Cuds


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class CeleryWorkflowSession(SimWrapperSession):
    """SimPhoNy-wrapper session for Celery-Chains and Worklows"""

    def __init__(
        self,
        input_uuid: "UUID",
        engine: CeleryWorkflowEngine = None,
        logging_id: str = None,
    ) -> None:
        """Initalite the session."""
        if not engine:
            engine = CeleryWorkflowEngine(input_uuid, logging_id=logging_id)
        super().__init__(engine=engine)

    # OVERRIDE
    def __str__(self):
        return "CeleryWorkflowSession"

    # OVERRIDE
    def _run(self, root_cuds_object) -> str:
        """Run the wrapper session."""
        return self._engine.run()

    # OVERRIDE
    def _apply_added(self, root_obj, buffer) -> None:
        """Apply scans of added cuds in buffer."""
        for obj in buffer.values():
            if obj.is_a(emmo.Workflow):
                message = """Found %s. Will search for workflow steps
                    and related workers on the platform."""
                logger.info(message, obj)
                self._scan_for_neighbours(obj)
        if not self._engine.tasks:
            message = """Did not find any workflow steps with
            complementary workers. Will scan for single
            object of type %s."""
            logger.info(message, emmo.Calculation)
            self._scan_for_single_calc(buffer)
        else:
            message = """Scan for workflow steps complete.
            Identified the following chain of workers: %s"""
            logger.info(message, self._engine.tasks)

    def _scan_for_neighbours(self, obj: "Cuds", step: str = "first") -> None:
        """Scan the input cuds for neighbour-tasks"""
        neighbour = obj.get(rel=emmo[f"hasSpatial{step.title()}"])
        if not neighbour:
            message = f"""Did not find {step} task in the chain of calculations in {obj}.
            Workflow chain has ended here."""
            logger.info(message)
        else:
            neighbour = neighbour.pop()
            self._get_worker_mapping(neighbour)
            self._scan_for_neighbours(neighbour, step="next")

    def _scan_for_single_calc(self, buffer: "Dict[Any, Cuds]") -> None:
        """Find single calculations to be run in the buffer."""
        for obj in buffer.values():
            if obj.is_a(emmo.Calculation):
                self._get_worker_mapping(obj)
        if not self._engine.tasks:
            message = """Did not find any calculations with
            complementary workers."""
            raise TypeError(message)
        message = """Found additional workers %s in the buffer,
        but will ignored because not part of a workflow chain."""
        logger.info(message, self._engine.tasks)

    def _get_worker_mapping(self, calculation: "Cuds") -> None:
        mapping = self._scan_worker_mapping(calculation)
        if not mapping:
            message = f"""Task in the chain of calculations is not properly
            mapped to an existing worker {calculation}."""
            logger.info(message)
        else:
            self._engine.add_task((calculation, mapping))

    def _scan_worker_mapping(self, calculation: "Cuds") -> "Optional[str]":
        response = []
        for superclass, mapping in self.worker_mapping.items():
            if calculation.is_a(superclass):
                response.append(mapping)
        if len(response) > 1:
            raise ValueError(
                f"More than 1 {calculation.oclass} found in worker mapping!"
            )
        if response:
            response = response.pop()
        return response

    # OVERRIDE
    def _apply_deleted(self, root_obj, buffer) -> None:
        """Apply functions for updated-buffer."""

    def _apply_updated(self, root_obj, buffer) -> None:
        """Apply functions for deleted-buffer."""

    # OVERRIDE
    def _load_from_backend(self, uids, expired=None) -> "Cuds":
        """Load a cuds from backend."""
        for uid in uids:
            if uid in self._registry:
                yield self._registry.get(uid)
            else:
                yield None

    @property
    def settings(cls) -> "BaseSettings":
        """Return the settings from the engine."""
        return cls._engine.settings

    @property
    def result(cls) -> "Dict[str, Any]":
        """Return the final result of the workflow"""
        return cls._engine.result

    @property
    def worker_mapping(cls) -> "Dict[str, Any]":
        """Return the mappings of the ontology class and worker name"""
        return cls._engine.worker_mapping

    @property
    def workers(cls) -> "List[str]":
        """Return the list of workers to pass the knowledge graph in a chain."""
        return cls._engine.workers



"""Pydantic Models for FastAPI-celery"""
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Literal, Optional
from uuid import UUID

from pydantic import BaseModel, Field


class RemoteWorker(str):
    """Identifier of the remote celery worker."""


class RemoteTaskName(str):
    """Task name which can be executed on remote worker."""


class ModelName(str):
    """Model name which can be instanciated via pydantic."""


class Registry(Dict):
    """Registry with available tasks available on remote workers."""


class TaskStatus(str, Enum):
    """Status of the task."""

    PENDING = "PENDING"
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"
    REVOKED = "REVOKED"


class TransformationStatus(str, Enum):
    """Status of the task."""

    CREATED = "CREATED"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    STOPPED = "STOPPED"
    RUNNING = "RUNNING"


task_to_transformation_map = {
    TaskStatus.PENDING: TransformationStatus.RUNNING,
    TaskStatus.SUCCESS: TransformationStatus.COMPLETED,
    TaskStatus.FAILURE: TransformationStatus.FAILED,
    TaskStatus.REVOKED: TransformationStatus.STOPPED,
}


UpdateTaskStates = Literal[TransformationStatus.RUNNING, TransformationStatus.STOPPED]


class TaskStatusModel(BaseModel):
    """Data model of the task status"""

    status: TaskStatus = Field(..., description="Status of the remote task.")
    state: TransformationStatus = Field(..., description="State of the remote task.")
    id: UUID = Field(..., description="UUID of the submitted task.")
    args: Optional["List[Any]"] = Field(
        ..., description="Arguments passed during the task submission."
    )
    kwargs: Optional["Dict[Any, Any]"] = Field(
        ..., description="Keyword arguments passed during the task submission."
    )
    traceback: Optional[str] = Field(
        ..., description="Traceback message about potential errors."
    )
    date_done: Optional[datetime] = Field(
        ..., description="Datetime when the submitted job finished."
    )


class TaskKillModel(TaskStatusModel):
    """Response model of the API when process of the task is killed."""

    message: str = Field(
        ...,
        description="""Human readable message with info
        if killing of task was successful.""",
    )


class TaskResultModel(BaseModel):
    """Data model of the task result"""

    id: UUID = Field(..., description="UUID of the submitted task.")
    parameters: Optional[Any] = Field(
        ..., description="Result of the task forwarded by celery-worker."
    )
    traceback: Optional[str] = Field(
        ..., description="Traceback message about potential errors."
    )
    date_done: Optional[datetime] = Field(
        ..., description="Datetime when the submitted job finished."
    )


class TaskResponseModel(BaseModel):
    """Response of the API with response to the task status"""

    message: str = Field(
        ..., description="Human readable message as response from request."
    )
    metadata: TaskStatusModel = Field(
        ..., description="Metadata about the task at the time of the request."
    )


class RegisteredTaskModel(BaseModel):
    """Response of the API data of the task status."""

    message: str = Field(
        ..., description="Human readable message as response from request."
    )
    registered_tasks: Dict[RemoteWorker, List[RemoteTaskName]] = Field(
        ..., description="Registry with tasks available on remote workers."
    )


class TaskCreateModel(BaseModel):
    """Response of the API when instanciating a data model."""

    id: str = Field(..., description="UUID of the instanciated data model in the cache")


class RegisteredModels(BaseModel):
    """Response of the API data of the task status."""

    message: str = Field(
        ..., description="Human readable message as response from request."
    )
    registered_models: List[ModelName] = Field(
        ..., description="Registry with data models available."
    )


class SubmissionBody(BaseModel):
    """Body of the workflow request."""

    state: UpdateTaskStates

    format: str = Field(
        "turtle", description="Format of the RDF-file to be consumed by the worker."
    )

    class Config:
        """Pydantic configuration for submission body"""

        schema_extra = {
            "example": {
                "state": "RUNNING",
                "format": "turtle",
            }
        }


class UploadDataResponse(BaseModel):
    """Body of the data upload response"""

    id: UUID = Field(
        ..., description="UUID of the data which was uploaded to the cache."
    )

    last_modified: str = Field(..., description="created time of the data.")


class InfoType(str, Enum):
    """Types of information to be returned from app"""

    SCHEMA = "Schema"
    EXAMPLE = "Example"


class UploadNotEnabledError(Exception):
    """Upload not permitted since not enabled by admin."""



"""Module for Celery-workflow engine."""
import logging
import tempfile
from typing import TYPE_CHECKING

from celery import Celery, signals
from celery.result import allow_join_result

from osp.core.namespaces import emmo, get_entity
from osp.core.session import CoreSession
from osp.core.utils import export_cuds, import_cuds
from osp.settings import AppConfig
from osp.utilities.load import get_download, get_upload

if TYPE_CHECKING:
    from typing import UUID, Any, Dict, List, Optional, Tuple

    from osp.core.cuds import Cuds
    from osp.core.ontology import OntologyClass


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


@signals.setup_logging.connect
def on_setup_logging(**kwargs):  # pylint: disable=unused-argument
    """Specify logging handler for tasks in celery worker."""
    # Get the root logger
    celery_logger = logging.getLogger()

    # Set the logging level to INFO
    celery_logger.setLevel(logging.INFO)

    # Add a console handler
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    console_handler.setFormatter(formatter)
    celery_logger.addHandler(console_handler)


class CeleryWorkflowEngine:
    """Class definition for Celery-workflow engine."""

    def __init__(
        self,
        input_uuid: "UUID",
        app: Celery = None,
        settings: AppConfig = None,
        logging_id: str = None,
    ) -> None:
        """Initalize the Celery-Workflow engine."""
        if not settings:
            settings = AppConfig()
        self._settings = settings
        if not app:
            address = self._settings.get_redis_address()
            app = Celery(broker=address, backend=address)
        self._logging_id: str = logging_id
        self._app: Celery = app
        self._input_uuid: "UUID" = input_uuid
        self._task_id: "Optional[UUID]" = None
        self._tasks: List = []
        self._results: List = []

    def add_task(self, mapping: "Tuple[Cuds, str]") -> None:
        """Add the name of an existing worker and the
        calculation type to batch."""
        self._tasks.append(mapping)

    def run(self) -> str:
        """Run the celery workflow engine"""
        return self._create_pipeline()

    # define a function to execute a task on a remote Celery worker
    def _execute_task(self, worker_name: str, task_id: str) -> "str":
        """Send a task to the Celery-app"""
        task = self._app.send_task(
            worker_name,
            kwargs={"cache_key": task_id, "task_id": self._logging_id},
            queue=worker_name,
        )
        with allow_join_result():
            result = task.get()
        self._results.append((worker_name, result))
        return result["cache_meta"]

    # define a function to create a chain of tasks based on the list of workers
    def _create_pipeline(self) -> str:
        """Create a list of tasks to execute in a chain"""
        uuid = self._input_uuid
        for calculation_cuds, worker_name in self._tasks:
            uuid = self._make_output_mapping(calculation_cuds, uuid)
            uuid = self._execute_task(worker_name, uuid)
        return uuid

    def _make_output_mapping(self, calculation: "Cuds", uuid: "UUID") -> "UUID":
        mappings = self._scan_output_mapping(calculation)
        if mappings:
            logger.info(
                """Found a match for input mapping for %s
                cue to outputs from a previous calculation: %s""",
                calculation,
                mappings,
            )
            cuds_file = get_download(uuid, as_file=True)
            core_session = CoreSession()
            import_cuds(cuds_file, session=core_session)
            query = core_session.load_from_iri(str(calculation.iri))
            current = query.first()  # pylint: disable=no-member
            if not current:
                raise ValueError(
                    f"Current calculation `{calculation}` not found in graph with object-id {uuid}"
                )
            for mapping in mappings:
                previous = mapping["previous"].iri
                output = mapping["output"].iri
                result = core_session.sparql(
                    f"""select ?output where {{
                        ?previous rdf:type <{previous}> .
                        ?previous <{emmo.hasOutput.iri}> ?output .
                        ?output rdf:type <{output}> .
                    }}"""
                )
                for row in result(output="cuds"):
                    current.add(row["output"], rel=emmo.hasInput)
            with tempfile.NamedTemporaryFile(suffix=".ttl", delete=False) as file:
                export_cuds(core_session, file.name)
                uuid = get_upload(file, uuid=uuid)
        return uuid

    def _scan_output_mapping(
        self, calculation: "Cuds"
    ) -> "List[Dict[str, OntologyClass]]":
        response = []
        for superclass, mapping in self.output_mapping.items():
            if calculation.is_a(superclass):
                response.append(mapping)
        if response:
            response = response.pop()
        return response

    @property
    def settings(cls):
        """Return the engine settings."""
        return cls._settings

    @property
    def worker_mapping(cls) -> "Dict[str, Any]":
        """Return the mappings of the ontology class and worker name"""
        return {
            get_entity(mapping.split(":")[0].strip()): mapping.split(":")[-1]
            for mapping in cls.settings.worker_mapping.split("|")
        }

    @property
    def output_mapping(cls) -> "Dict[str, Any]":
        "Output mapping for previous calculations for next calculation"
        mapping = {}
        for entry in cls.settings.output_mapping.split("|"):
            entities = [get_entity(entity.strip()) for entity in entry.split(":")]
            key = entities[0]
            values = [{"previous": entities[1], "output": entities[2]}]
            if key in mapping:
                mapping[key] += values
            else:
                mapping[key] = values
        return mapping

    @property
    def result(cls) -> "Dict[str, Any]":
        """Return the final result of the workflow"""
        return {
            f"{n}_{worker_name}": result["cache_raw"]
            for n, (worker_name, result) in enumerate(cls.results)
        }

    @property
    def results(cls) -> "List[Dict[str, Any]]":
        """Return the results from the celery engine."""
        return cls._results

    @property
    def tasks(cls) -> "List[str]":
        """Return the list of tasks to pass the knowledge graph in a chain."""
        return cls._tasks



"""Pydantic settings for FastAPI and Celery."""

from typing import Optional

from fastapi_plugins import RedisSettings
from pydantic import Field, SecretStr


class AppConfig(RedisSettings):
    """Main configuration for based on redis-settings"""

    worker_name: str = Field(
        "simphony-workflows",
        description="Name of the worker displayed in the services",
    )

    output_mapping: str = Field(
        """emmo.MesoscopicCalculation:emmo.BindingSites:crystallography.UnitCell |
        emmo.MesoscopicCalculation:emmo.ProcessSearch:emmo.ClusterExpansion |
        emmo.MesoscopicCalculation:emmo.ProcessSearch:emmo.ChemicalReactionMechanism"""
    )

    worker_mapping: str = Field(
        "emmmo:ContinuumCalculation:simphony-catalyticfoam",
        description="""Mapping between the label of the entity in EMMO representing
        a calculation e.g. `ContinuumCalculation` for the CO-Usecase in CatalyticFoam
        and the name of the worker executing the wrapper, e.g. `simphony-catalyticfoam`.
        Key-value sepator is `:` whereas the item-separator is `|`.""",
    )

    schemas: str = Field(
        "osp.models.co_catalyticfoam:COCatalyticFOAMModel",
        description="""Models available to be instanciated via Pydantic.
        Each item is a python class to be imported from a modle.
        e.g. `COCatalyticFOAMModel` for the CO-Usecase in CatalyticFoam
        with `osp.models.co_catalyticfoam` as respective Python-modle.
        Key-value sepator is `:` whereas the item-separator is `|`.""",
    )

    authentication_dependencies: bool = Field(
        False, description="If authentication should be enabled or not."
    )

    wrapper_name: str = Field(
        "osp.wrappers.celery_workflow_wrapper:CeleryWorkflowSession",
        description="""SimWrapperSession class to be loaded from a python module
        for wrapping to the celery backend. The regex is: mypackage.mymodule:MyClass""",
    )

    minio_user: SecretStr = Field(..., description="User to MinIO instance.")
    minio_password: SecretStr = Field(..., description="Password to MinIO instance.")
    minio_endpoint: str = Field(
        "localhost:9000",
        description="Resolvable endpoint to contact MinIO instance.",
    )
    external_hostname: Optional[str] = Field(
        None, description="Resolvable hostname to the outside world."
    )

    enable_upload: Optional[bool] = Field(
        False,
        description="""Whether direct data upload into
        the cache should be enabled or not""",
    )

    class Config:
        """Pydantic config for FastAPI-celery settings"""

        env_prefix = "REAXPRO_"



"""Helper functions for OSP-utilities."""
import io
import os
from uuid import uuid4

from minio import Minio
from urllib3.response import HTTPResponse

from .exceptions import MinioConnectionError, MinioDownloadError


def _get_upload(filepath: str, uuid: str, minio_client: Minio) -> str:
    """Helper function for `depends_upload` and `get_upload`."""
    suffix = os.path.splitext(filepath)[-1]
    # Generate a unique UUID as the object key
    object_key = uuid or str(uuid4())

    # Upload the file to MinIO
    try:
        if not minio_client.bucket_exists(object_key):
            minio_client.make_bucket(object_key)
        minio_client.fput_object(
            object_key,
            object_key,
            filepath,
            metadata={"suffix": suffix},
        )
    except Exception as err:
        raise MinioConnectionError(err.args) from err
    return object_key


def _get_download(uuid: str, minio_client: Minio) -> HTTPResponse:
    """Helper function for `depends_download` and `get_download`."""
    file_data = io.BytesIO()
    if not minio_client.bucket_exists(uuid):
        raise MinioDownloadError("Bucket does not exist.")
    try:
        response = minio_client.get_object(
            uuid,
            uuid,
            file_data,
        )
    except Exception as err:
        raise MinioConnectionError(err) from err
    return response



"""osp-utilities for boto3"""
import boto3
from botocore.client import BaseClient

from osp.settings import AppConfig


def get_boto3() -> BaseClient:
    """Helper function for returning S3-client using MinIO."""
    settings = AppConfig()
    return boto3.client(
        "s3",
        endpoint_url=f"http://{settings.minio_endpoint}",
        aws_access_key_id=settings.minio_user.get_secret_value(),
        aws_secret_access_key=settings.minio_password.get_secret_value(),
    )



"""Celery tasks for FastAPI"""

import logging
import tempfile
from importlib import import_module
from typing import TYPE_CHECKING

from celery import Celery, current_task, signals

from osp.app.s3 import get_s3handler
from osp.core.cuds import Cuds
from osp.core.namespaces import cuba
from osp.core.utils import export_cuds, import_cuds
from osp.settings import AppConfig
from osp.utilities import get_download, get_upload

if TYPE_CHECKING:
    from typing import Callable


def get_wrapper_class(module: str) -> "Callable":
    """Get wrapper from import-specification"""
    module, classname = module.strip().split(":")
    return getattr(import_module(module), classname)


settings = AppConfig()
address = settings.get_redis_address()
celery = Celery(settings.worker_name, broker=address, backend=address)
celery.conf.CELERYD_HIJACK_ROOT_LOGGER = False


@signals.setup_logging.connect
def on_setup_logging(**kwargs):  # pylint: disable=unused-argument
    """Specify logging handler for tasks in celery worker."""
    # Get the root logger
    logger = logging.getLogger()

    # Set the logging level to INFO
    logger.setLevel(logging.INFO)

    # Add a console handler
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)


@celery.task(name=settings.worker_name)
def run_simulation(
    cache_key: str = None, task_id: str = None, store_tarball: bool = True
) -> str:
    """Run celery-workflow wrapper as celery-task."""

    # Configure the logging module
    task_id = task_id or current_task.request.id
    s3_handler = get_s3handler(task_id)
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    s3_handler.setFormatter(formatter)
    logging.basicConfig(level=logging.INFO)
    logging.root.addHandler(s3_handler)

    # download cuds
    logging.info("received cache_key %s", cache_key)
    cudspath = get_download(cache_key, as_file=True)

    # run wrapper
    session_class = get_wrapper_class(settings.wrapper_name)
    with session_class(input_uuid=cache_key, logging_id=task_id) as session:
        wrapper = cuba.Wrapper(session=session)
        cuds = import_cuds(cudspath, session=session)
        if isinstance(cuds, list):
            wrapper.add(*cuds, rel=cuba.relationship)
        elif isinstance(cuds, Cuds):
            wrapper.add(cuds, rel=cuba.relationship)
        session.run()

    # upload Cuds
    with tempfile.NamedTemporaryFile(suffix=".ttl", delete=False) as file:
        export_cuds(session, file.name)
        meta_key = get_upload(file)

    # get tarball keys
    if store_tarball:
        with open(session.tarball, "rb") as tar:
            tar_key = get_upload(tar)
    else:
        tar_key = session.result

    return {"cache_meta": meta_key, "cache_raw": tar_key}



"""osp-utilities for up/downloads with minio"""
import tempfile
from typing import TextIO, Union

from urllib3.response import HTTPResponse

from .helper import _get_download, _get_upload
from .minio import get_minio


def get_upload(file: Union[str, TextIO], uuid: str = None) -> str:
    """Upload file with minio client and return upload-id without `Depends`."""
    if hasattr(file, "filename"):
        file = file.filename
    elif hasattr(file, "name"):
        file = file.name
    minio_client = get_minio()
    return _get_upload(file, uuid, minio_client)


def get_download(uuid: str, as_file=False) -> Union[HTTPResponse, str]:
    """Return `io.BytesIO` from uuid through minio client without `Depends`."""
    minio_client = get_minio()
    response = _get_download(uuid, minio_client)
    if as_file:
        suffix = response.headers.get("x-amz-meta-suffix")
        with tempfile.NamedTemporaryFile(suffix=suffix, delete=False) as temp:
            temp.write(response.data)
            response = temp.name
    return response



"""Commandline tool for starting the reaxpro platform"""

import atexit
import os
import subprocess  # nosec
import sys
from typing import Dict, List, Optional

import click
import uvicorn
import yaml
from pydantic import BaseModel


class WorkerModel(BaseModel):
    """Pydantic model for all workers orechestered by head worker"""

    name: str
    wrapper: str
    extra_env: Optional[Dict[str, str]]


class HeadModel(BaseModel):
    """Pydantic model for head worker"""

    name: str
    mapping: str


class BrokerModel(BaseModel):
    """Pydantic model message broker details"""

    type: str
    host: str
    port: int
    db: int


class MinioModel(BaseModel):
    """Pydantic model minio object store details"""

    user: str
    password: str
    endpoint: str


class FastAPIModel(BaseModel):
    """Pydantic model FastAPI app details"""

    host: str
    port: int
    schemas: str


class AppConfigModel(BaseModel):
    """Pydantic model overall app config"""

    fastapi: Optional[FastAPIModel]
    minio: MinioModel
    broker: BrokerModel
    head: Optional[HeadModel]
    workers: Optional[List[WorkerModel]]


class ReaxProConfigModel(BaseModel):
    """Pydantic model overall app config defined by `reaxpro`-submodel"""

    reaxpro: AppConfigModel


@click.command()
@click.option(
    "--config",
    default="cmdline-config.yml",
    type=click.Path(exists=True, dir_okay=False),
    multiple=False,
    help="Config starting platform",
    envvar="REAXPRO_CMDLINE_CONFIG",
)
def start_platform(config):
    """Run reaxpro platform with respect to the passed configuration"""
    # define colors for echo
    green = "\033[92m"
    red = "\033[91m"
    reset = "\033[0m"

    # open the yml file
    with open(config, "r", encoding="utf-8") as stream:
        try:
            config = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            raise ValueError(
                f"Could not parse YML due to following error: {exc}"
            ) from exc

    # parse and validate model due to config
    try:
        model = ReaxProConfigModel(**config)
    except Exception as execp:
        raise ValueError(f"Invalid configuration: {execp}") from execp

    # export minio variables
    os.environ["REAXPRO_MINIO_USER"] = model.reaxpro.minio.user
    os.environ["REAXPRO_MINIO_PASSWORD"] = model.reaxpro.minio.password
    os.environ["REAXPRO_MINIO_ENDPOINT"] = model.reaxpro.minio.endpoint

    # export redis variables
    os.environ["REAXPRO_REDIS_TYPE"] = model.reaxpro.broker.type
    os.environ["REAXPRO_REDIS_HOST"] = model.reaxpro.broker.host
    os.environ["REAXPRO_REDIS_PORT"] = str(model.reaxpro.broker.port)
    os.environ["REAXPRO_REDIS_DB"] = str(model.reaxpro.broker.db)

    # define processes varaible
    reaxpro_processes = {}

    # define cleanup function when tool is exciting
    def cleanup():
        """Terminate worker processes."""
        for name, process in reaxpro_processes.items():
            click.echo(
                f"""{green}INFO{reset}:\tTerminating"""
                f"""worker with name `{name}` and pid `{process.pid}`"""
            )
            process.terminate()
            process.wait()

    # register cleanup
    atexit.register(cleanup)

    if model.reaxpro.head:
        click.echo(f"{green}INFO{reset}:\tStarting head worker with Celery...")
        reaxpro_processes["headworker"] = _run_head(model.reaxpro.head)

    if model.reaxpro.workers:
        for worker_model in model.reaxpro.workers:
            click.echo(
                f"""{green}INFO{reset}:\t"""
                f"""Starting {worker_model.name}-worker with Celery..."""
            )
            reaxpro_processes[worker_model.name] = _run_worker(worker_model)

    if model.reaxpro.fastapi:
        from .main import app  # pylint: disable=import-outside-toplevel

        # fastapi variables
        os.environ["REAXPRO_SCHEMAS"] = model.reaxpro.fastapi.schemas

        # start fastapi
        click.echo(f"{green}INFO{reset}:\tStarting FastAPI...")
        uvicorn.run(
            app,
            host=model.reaxpro.fastapi.host,
            port=model.reaxpro.fastapi.port,
            log_level="debug",
        )

    elif reaxpro_processes:
        # grab any of the workers and let wait for the process
        workers = ",".join([f"`{worker}`" for worker in reaxpro_processes])
        click.echo(
            f"{green}INFO{reset}:\t Workers are ready for receiving tasks: {workers}"
        )
        process = next(iter(reaxpro_processes.values()))
        process.wait()

    else:
        click.echo(
            f"{red}ERROR{reset}:\tNo processes defined, will shutdown application..."
        )
        sys.exit(1)


def _run_head(model: HeadModel) -> subprocess.Popen:
    command = [
        sys.executable,
        "-m",
        "celery",
        "-A",
        "osp.app.tasks:celery",
        "worker",
        "-Q",
        model.name,
        "-n",
        model.name,
    ]
    env = os.environ.copy()
    env["REAXPRO_WORKER_NAME"] = model.name
    env["REAXPRO_WORKER_MAPPING"] = model.mapping
    return _run_command(command, env)


def _run_worker(model: WorkerModel) -> subprocess.Popen:
    command = [
        sys.executable,
        "-m",
        "celery",
        "-A",
        "osp.app.tasks:celery",
        "worker",
        "-Q",
        model.name,
        "-n",
        model.name,
    ]
    env = os.environ.copy()
    env["REAXPRO_WORKER_NAME"] = model.name
    env["REAXPRO_WRAPPER_NAME"] = model.wrapper
    if model.extra_env:
        for key, value in model.extra_env.items():
            env[key] = value
    return _run_command(command, env)


def _run_command(command: List[str], env: Dict[str, str]) -> subprocess.Popen:
    return subprocess.Popen(
        command, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
    )  # nosec


if __name__ == "__main__":
    start_platform()  # pylint: disable=no-value-for-parameter



"""Setup class for Reaxpro-workflow-service"""
import warnings

from setuptools import setup

setup()

VERSION: str = "v1.0.0"
BASE = "https://raw.githubusercontent.com/simphony/reaxpro-framework-ontology"
URL = f"{BASE}/{VERSION}/reaxpro-inferred.ttl"
CONTENT = f"""identifier: reaxpro
ontology_file: '{URL}'
reference_by_label: True
format: ttl
namespaces:
    emmo: http://emmo.info/emmo
    crystallography: http://emmo.info/domain-crystallography/crystallography
active_relationships:
    - http://emmo.info/emmo#EMMO_8c898653_1118_4682_9bbf_6cc334d16a99
    - http://emmo.info/emmo#EMMO_60577dea_9019_4537_ac41_80b0fb563d41
    - http://emmo.info/emmo#EMMO_8785be5a_2493_4b12_8f39_31907ab11748
    - http://emmo.info/emmo#EMMO_52d08d7d_e9e4_43e5_8508_d353e7e3a23a
default_relationship: http://emmo.info/emmo#EMMO_17e27c22_37e1_468c_9dd7_95e137f73e7f
"""


def install_ontology():
    """Generates and installs the reaxpro ontology."""
    from tempfile import NamedTemporaryFile

    from osp.core.ontology.installation import OntologyInstallationManager

    with NamedTemporaryFile("w", suffix=".yml", delete=False) as yml_file:
        yml_file.write(CONTENT)

    manager = OntologyInstallationManager()
    manager.install_overwrite(yml_file.name)


# install ontology
try:
    install_ontology()
except Exception as err:
    warnings.warn(f"Ontologies were not properly installed: {err}")



"""OSP-utilities"""

from .boto3 import get_boto3
from .load import get_download, get_upload
from .minio import get_minio

__all__ = [
    "get_download",
    "get_upload",
    "get_boto3",
    "get_minio",
]
